/* --COPYRIGHT--,BSD
 * Copyright (c) 2016, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
//*****************************************************************************
//         MSP430FR2311 - IR Reflection Sensing TI design firmware
//
// This application operates on TIDM-FRAM-IRREFLECTIONSENSING board to
// demonstrate IR reflection sensing function with FRAM based MSP430FR2311.
// Texas Instruments, Inc.
// Ver 0.1 July. 2016
// ******************************************************************************

#include <msp430.h>
#include <stdint.h>


//#define AC_PATH

// P2.0 for Buzzer
#define  BUZZER     BIT0
#define  BUZZERDIR  P2DIR
#define  BUZZEROUT  P2OUT
#define  BUZZERSEL  P2SEL0

// P2.6 for LED
#define  LED        BIT6
#define  LEDDIR     P2DIR
#define  LEDOUT     P2OUT

// P2.7 for IR LED
#define  IRLED      BIT7
#define  IRLEDDIR   P2DIR
#define  IRLEDOUT   P2OUT

// P1.1 for Button
#define  BUTTON     BIT1
#define  BUTTONIN   P1IN
#define  BUTTONOUT  P1OUT
#define  BUTTONIE   P1IE
#define  BUTTONIES  P1IES
#define  BUTTONREN  P1REN
#define  BUTTONIFG  P1IFG


#define RTC_CAL 50                          // VLO calibration time setting
#define INTERVAL_TIME 8                     // Wake up every 8 seconds
#define LED_INTERVAL 5                      // LED blink every 48 seconds

#define MCLK_FREQ_MHZ 1                     // MCLK = 1MHz

#define BUFFER_SIZE 4                       // Buffer for ADC samples
uint16_t dark_buffer[BUFFER_SIZE]={0};
uint16_t light_buffer[BUFFER_SIZE]={0};

uint16_t dark_average;                      // Average samples with IRLED off
uint16_t light_average;                     // Average samples with IRLED on
uint8_t  IR_detect_count = 0;               // Counter for IR reflection sensing
uint8_t  alarm_flag = 0;                    // Alarm status. 1 means IR sensed
uint8_t  alarm_disable_count = 0;           // Alarm enable. 0 means enable

uint8_t  sample_cnt = 0;                    // Indicate IR sensing status
uint8_t  test_button = 1;                   // Button input value
uint8_t  interrupt_edge;                    // Button interrupt edge
uint16_t rtc_cnt;                           // Variable for RTC setting
uint16_t counts;                            // Counts of VLO calibration
uint16_t cap_cnt1,cap_cnt2;                 // Timer capture values
uint8_t  led_cnt = 0;                       // Counter for LED blink interval

void setup(void);                           // Setup clock and GPIO
void sample(void);                          // ADC sampling
void average(void);                         // Calculate ADC average value
void cal_VLO(void);                         // VLO calibration
void init_GPIO(void);                       // Initialize GPIO
void enable_TIA(void);                      // Enable TRIamp
void enable_SAC(void);                      // Enable SAC(OPA)
void disable_TIA(void);                     // Disable TRIamp
void disable_SAC(void);                     // Disable SAC(OPA)
void alarm_on(void);                        // Turn on alarm
void alarm_off(void);                       // Turn off alarm
void Software_Trim();                       // Software Trim to get the best DCOFTRIM value


void main(void)
{
	WDTCTL = WDTPW + WDTHOLD;               // Stop watchdog timer
	setup();                                // Call init routine

	while(1)
	{

		cal_VLO();                          // calibrate VLO


		rtc_cnt = RTC_CAL * INTERVAL_TIME * 32768 / counts;
		RTCMOD = rtc_cnt -RTC_CAL*2 -1;     // setup RTC for 8 second interrupt
		RTCCTL = RTCSS__VLOCLK | RTCSR | RTCPS__10 | RTCIE; // Source = VLO, divided by 10

		sample();                           // ADC sampling
		average();                          // Calculate ADC average value

		if(alarm_disable_count > 0)         // Alarm is disabled
			alarm_disable_count--;          // Timing counter for alarm disable

		if ((alarm_disable_count == 0) && (light_average > (dark_average + 30)))
		{
			if (alarm_flag == 0)
			{
				IR_detect_count++;       // Increment detection counter
				IR_detect_count &= 0x03; // Only count values of up to 3

				if (IR_detect_count == 1)
				{
					rtc_cnt = rtc_cnt>>2;   // makes the sample period 2s
					RTCMOD = rtc_cnt -RTC_CAL*2 -1;
					RTCCTL |= RTCSR;
				}
				else if (IR_detect_count == 2)
				{
					rtc_cnt = rtc_cnt>>3;   // makes the sample period 1s
					RTCMOD = rtc_cnt -RTC_CAL*2 -1;
					RTCCTL |= RTCSR;
				}
				else if (IR_detect_count == 3)
				{
					alarm_flag = 1;         // Set alarm flag
				}
			}
		}
		else   // light_average <= dark_average+30
		{
			IR_detect_count = 0;         // Clear IR detect counter
		}

		if(alarm_flag == 1)                 // If IR present, turn on alarm
		{
			rtc_cnt = rtc_cnt>>4;           // makes the sample period 0.5s
			RTCMOD = rtc_cnt -RTC_CAL*2 -1;
			RTCCTL |= RTCSR;
			alarm_on();
			__bis_SR_register(LPM3_bits);   // Enter LPM3
		}
		else
		{
            if(led_cnt < LED_INTERVAL-1)    // LED blink every 48 seconds
            {
            	led_cnt++;
            }
            else
            {
            	led_cnt = 0;
            	LEDOUT |= LED;              // Blink visible LED
			    __delay_cycles(10000);      // 10ms
			    LEDOUT &= ~LED;
            }
			P1OUT ^= BIT0;                  // For test only, test timing accuracy
			__bis_SR_register(LPM4_bits);   // Enter LPM4
		}
	}
}


void setup(void)
{
	uint16_t i;

	  __bis_SR_register(SCG0);                // Disable FLL
	  CSCTL3 = SELREF__REFOCLK;               // Set REFO as FLL reference source
	  CSCTL1 = DCOFTRIMEN_1 | DCOFTRIM0 | DCOFTRIM1 | DCORSEL_0;// DCOFTRIM=3, DCO Range = 1MHz
	  CSCTL2 = FLLD_0 + 30;                   // DCODIV = 1MHz
	  __delay_cycles(3);
	  __bic_SR_register(SCG0);                // Enable FLL
	  Software_Trim();                        // Software Trim to get the best DCOFTRIM value
	  CSCTL4 = SELMS__DCOCLKDIV | SELA__REFOCLK; // set default REFO(~32768Hz) as ACLK source, ACLK = 32768Hz
	                                             // default DCODIV as MCLK and SMCLK source
	init_GPIO();

	for(i=0;i<10;i++)                       // Blink LED to indicate reset
	{
		LEDOUT ^= LED;
		__delay_cycles(80000);
	}
	LEDOUT &= ~LED;
}

void init_GPIO(void)
{
	// Port output low
	P1OUT = 0x00;
	P1DIR |= 0xff;
	P2OUT = 0x00;
	P2DIR |= 0xff;

	// Setup Red LED as off (output low)
	LEDOUT &= ~LED;
	LEDDIR |= LED;

	// Setup IR LED pin as output low
	IRLEDOUT &= ~IRLED;
	IRLEDDIR |= IRLED;

	// Setup Buzzer pin as output low
	BUZZEROUT &= ~BUZZER;
	BUZZERDIR |= BUZZER;

	// BUTTON as interrupt
	BUTTONIE |= BUTTON;                     // Enable BUTTON interrupt
	BUTTONIES |= BUTTON;                    // H to L transition for interrupt
	BUTTONREN |= BUTTON;                    // Enable internal pull resistor
	BUTTONOUT |= BUTTON;                    // Set internal pull resistor as pull-up
	BUTTONIFG &= ~BUTTON;                   // Clear any pending ISRs

	PM5CTL0 &= ~LOCKLPM5;                   // Disable the GPIO power-on default high-impedance mode
	                                        // to activate previously configured port settings
}

void enable_TIA(void)
{
	//Select P1.5 P1.7 as TIA fucntion
	P1SEL0 |= BIT5 + BIT7;
	P1SEL1 |= BIT5 + BIT7;

	P1OUT |= BIT6;
	P1DIR |= BIT6;                          // Generate bias current for TIA

	TRI0CTL = TRIPM + TRIEN;                // Enable TIA, Select low speed and low power mode
}

void enable_SAC(void)
{
	//Select P1.2 P1.3 P1.4 as SAC function
	P1SEL0 |= BIT2 + BIT3 + BIT4;
	P1SEL1 |= BIT2 + BIT3 + BIT4;

	SAC0OA |= PSEL1;                        // Select TIA output as SAC positive input source
	SAC0OA |= NMUXEN + PMUXEN;              // Enable negative and positive input
	SAC0OA |= OAPM;                         // Select low speed and low power mode
	SAC0OA |= SACEN + OAEN;                 // Enable SAC and OA
}

void disable_TIA(void)
{
	TRI0CTL = 0;                            // Disable TIA
	//Select P1.5 P1.7 as GPIO fucntion
	P1SEL0 &= ~(BIT5 + BIT7);
	P1SEL1 &= ~(BIT5 + BIT7);
	P1OUT &= ~BIT6;

}

void disable_SAC(void)
{
	SAC0OA = 0;                             // Disable SAC
	//Select P1.2 P1.3 P1.4 as GPIO function
	P1SEL0 &= ~(BIT2 + BIT3 + BIT4);
	P1SEL1 &= ~(BIT2 + BIT3 + BIT4);
}

void alarm_on(void)
{
	LEDOUT |= LED;                          // Turn on visible LED
	// Enable TB1 output for alarm cadence
	TB1CCR0 = 33;                           // PWM Period/2
	TB1CCR1 = 16;                           // TBCCR1 PWM duty cycle
	TB1CCTL1 = OUTMOD_6;                    // TBCCR1 toggle/set
	TB1CTL = TBSSEL__ACLK | MC_3;           // ACLK, up-down mode
	BUZZERSEL |= BUZZER;                    // Enable BUZZER pin for TB1 output
}

void alarm_off(void)
{
	TB1CTL = 0;                             // Disable Buzzer
	TB1CCTL1 = 0;
	LEDOUT &= ~LED;                         // Turn off visible LED
}

void cal_VLO(void)
{
	cap_cnt1 = 0;                           // First capture counter
	cap_cnt2 = 0;                           // Second capture counter

	RTCMOD = RTC_CAL -1;                    // Ouput RTC pulse every RTC_CAL VLO clock cycles
    RTCCTL = RTCSS__VLOCLK | RTCSR | RTCPS__10;
                                        	// Timer0_B3 Setup
	TB0CCTL0 |= CM_1 | CCIS_0 | CAP | CCIE; // Capture on RTC output
	                                        // Capture rising edge,
	                                        // Use CCI0A=RTC,
	                                        // Enable capture mode,
	TB0CTL |= TBSSEL__ACLK | MC_2 | TBCLR;  // Use ACLK as clock source, clear TB0R
                                          	// Start timer in continuous mode
	__bis_SR_register(LPM3_bits + GIE);

	__no_operation();

	__bis_SR_register(LPM3_bits);           // Calibrate in LPM3 with REFO

    counts = cap_cnt2 - cap_cnt1;           // Get VLO calibration counter value

    TB0CTL = 0;                             // Stop timer
    TB0CCTL0 = 0;
    RTCCTL = 0;                             // Stop RTC
}


void sample(void)
{
	sample_cnt=0;

	enable_TIA();                           // Enable TIA
	enable_SAC();                           // Enable SAC(OPA)

	// Configure ADC to repeat-single-channel mode
	ADCCTL0 = ADCSHT_1 | ADCMSC | ADCON;    // ADCON
	ADCCTL1 = ADCSHP | ADCCONSEQ_2;         // Rpt single ch;
	ADCCTL2 = ADCRES;                       // 10-bit conversion results
#ifdef AC_PATH
	ADCMCTL0 = ADCINCH_4 | ADCSREF_0;       // A4 ADC input select = SAC output; Vref = DVCC
#else
	ADCMCTL0 = ADCINCH_3 | ADCSREF_0;       // A3 ADC input select = SAC output; Vref = DVCC
//	ADCMCTL0 = ADCINCH_5 | ADCSREF_0;       // A5 ADC input select = TIA output; Vref = DVCC
#endif

	__delay_cycles(80);                     // Wait for SAC&TIA setup

	ADCCTL0 |= ADCENC | ADCSC;              // Sampling and conversion start

	while(sample_cnt < BUFFER_SIZE)
	{
		if(ADCIFG&ADCIFG0)
		{
			dark_buffer[sample_cnt] = ADCMEM0;
			sample_cnt++;
		}
	}
	sample_cnt = 0;

	IRLEDOUT |= IRLED;                      // Turn on IRLED
	__delay_cycles(10);

	while(sample_cnt < BUFFER_SIZE)
	{
		if(ADCIFG&ADCIFG0)
		{
			light_buffer[sample_cnt] = ADCMEM0;
			sample_cnt++;
		}
	}
	IRLEDOUT &= ~IRLED;                     // Turn off IRLED
	sample_cnt = 0;

	ADCCTL0 &= ~ (ADCENC + ADCSC);          // Disable ADC
	ADCCTL0 &= ~ADCON;

	disable_SAC();                          // Disable SAC
	disable_TIA();                          // Disable TIA
}


void average(void)
{
	uint16_t i;

	dark_average = 0;                      // dark:  IR LED off
	light_average = 0;                     // light: IR LED on

	for (i = 0; i < BUFFER_SIZE; i++)
	{
		dark_average += dark_buffer[i];
		light_average += light_buffer[i];
		dark_buffer[i] = 0;
		light_buffer[i] = 0;
	}
	dark_average = (dark_average >> 2);     // BUFFER_SIZE = 4
	light_average = (light_average >> 2);   // Calculate ADC average value
}

void Software_Trim()
{
    unsigned int oldDcoTap = 0xffff;
    unsigned int newDcoTap = 0xffff;
    unsigned int newDcoDelta = 0xffff;
    unsigned int bestDcoDelta = 0xffff;
    unsigned int csCtl0Copy = 0;
    unsigned int csCtl1Copy = 0;
    unsigned int csCtl0Read = 0;
    unsigned int csCtl1Read = 0;
    unsigned int dcoFreqTrim = 3;
    unsigned char endLoop = 0;

    do
    {
        CSCTL0 = 0x100;                         // DCO Tap = 256
        do
        {
            CSCTL7 &= ~DCOFFG;                  // Clear DCO fault flag
        }while (CSCTL7 & DCOFFG);               // Test DCO fault flag

        __delay_cycles((unsigned int)3000 * MCLK_FREQ_MHZ);// Wait FLL lock status (FLLUNLOCK) to be stable
                                                           // Suggest to wait 24 cycles of divided FLL reference clock
        while((CSCTL7 & (FLLUNLOCK0 | FLLUNLOCK1)) && ((CSCTL7 & DCOFFG) == 0));

        csCtl0Read = CSCTL0;                   // Read CSCTL0
        csCtl1Read = CSCTL1;                   // Read CSCTL1

        oldDcoTap = newDcoTap;                 // Record DCOTAP value of last time
        newDcoTap = csCtl0Read & 0x01ff;       // Get DCOTAP value of this time
        dcoFreqTrim = (csCtl1Read & 0x0070)>>4;// Get DCOFTRIM value

        if(newDcoTap < 256)                    // DCOTAP < 256
        {
            newDcoDelta = 256 - newDcoTap;     // Delta value between DCPTAP and 256
            if((oldDcoTap != 0xffff) && (oldDcoTap >= 256)) // DCOTAP cross 256
                endLoop = 1;                   // Stop while loop
            else
            {
                dcoFreqTrim--;
                CSCTL1 = (csCtl1Read & (~DCOFTRIM)) | (dcoFreqTrim<<4);
            }
        }
        else                                   // DCOTAP >= 256
        {
            newDcoDelta = newDcoTap - 256;     // Delta value between DCPTAP and 256
            if(oldDcoTap < 256)                // DCOTAP cross 256
                endLoop = 1;                   // Stop while loop
            else
            {
                dcoFreqTrim++;
                CSCTL1 = (csCtl1Read & (~DCOFTRIM)) | (dcoFreqTrim<<4);
            }
        }

        if(newDcoDelta < bestDcoDelta)         // Record DCOTAP closest to 256
        {
            csCtl0Copy = csCtl0Read;
            csCtl1Copy = csCtl1Read;
            bestDcoDelta = newDcoDelta;
        }

    }while(endLoop == 0);                      // Poll until endLoop == 1

    CSCTL0 = csCtl0Copy;                       // Reload locked DCOTAP
    CSCTL1 = csCtl1Copy;                       // Reload locked DCOFTRIM
    while(CSCTL7 & (FLLUNLOCK0 | FLLUNLOCK1)); // Poll until FLL is locked
}

// RTC interrupt service routine
#pragma vector=RTC_VECTOR
__interrupt void RTC_ISR(void)
{
	switch(__even_in_range(RTCIV,RTCIV_RTCIF))
	{
	case  RTCIV_NONE:   break;              // No interrupt
	case  RTCIV_RTCIF:                      // RTC Overflow
		__bic_SR_register_on_exit(LPM4_bits);// Exit LPM4
		break;
	default: break;
	}
}

// TimerB interrupt service routine
#pragma vector = TIMER0_B0_VECTOR
__interrupt void TIMER0_B0_ISR(void)
{
	if(cap_cnt1 ==0)
	  	cap_cnt1 = TB0CCR0;                 // First capture counter
	else
	  	cap_cnt2 = TB0CCR0;                 // Second capture counter
	TB0CCTL0 &= ~CCIFG;                     // Clear timer IFG
	__bic_SR_register_on_exit(LPM3_bits);   // Exit LPM3
}

// PORT1 interrupt service routine
#pragma vector=PORT1_VECTOR
__interrupt void P1_ISR(void)
{
	P1IFG = 0;                              // Clear Port1 IFG
	__delay_cycles(50000);                  // 50ms Debounce delay
	interrupt_edge = BUTTONIES & BUTTON;
	if(interrupt_edge == 2)                 // Check interrupt edge
	{
		if(alarm_flag == 1)                 // Short time press to turn off alarm
		{
			alarm_off();
			alarm_flag = 0;
			alarm_disable_count = 6;        // Disable alarm for 48 seconds
		}
		else
		{
		    __delay_cycles(2000000);
		    test_button = BUTTONIN & BUTTON; // Check whether test buttton is pressed more than 2 seconds
		    if(test_button == 0)             // Long time press to turn on alarm
			alarm_on();
		}
		BUTTONIES &= ~BUTTON;               // L to H transition for interrupt
	}
	else
	{
        alarm_off();                        // Alarm is turned off if button is released
	    BUTTONIES |= BUTTON;                // H to L transition for interrupt
	}
}
